import { IModel, ModelManager } from 'rev-models';
import { MongoClientOptions } from 'mongodb';
import { IBackend } from 'rev-models/lib/backends';
import { ModelOperationResult } from 'rev-models/lib/operations/operationresult';
import { ICreateMeta, IUpdateMeta, IRemoveMeta, IReadMeta, IExecMeta } from 'rev-models/lib/models/types';
import { ICreateParams, IUpdateParams, IRemoveParams, IReadParams, IExecParams } from '../../rev-models/lib/backends/backend';
/**
 * Configuration used to initialise the MongoDB connection for [[MongoDBBackend]]
 */
export interface IMongoDBBackendConfig {
    url: string;
    dbName: string;
    options?: MongoClientOptions;
}
/**
 * The MongoDB Collection used to store the next @AutoNumberField values
 */
export declare const AUTONUMBER_COLLECTION_NAME = "__revjs_autonumber";
/**
 * A RevJS Backend class that stores model data in MongoDB
 *
 * Usage example:
 *
 * ```ts
 * [[include:examples/src/using_backends/using_a_mongodb_backend.ts]]
 * ```
 */
export declare class MongoDBBackend implements IBackend {
    private config;
    private client;
    private db;
    /**
     * Backend constructor.
     *
     * **NOTE:** You must call the **[[MongoDBBackend.connect]]** method after constructing
     * the backend before attempting to store or retrieve any data.
     * @param config MongoDB Connection Information
     */
    constructor(config: IMongoDBBackendConfig);
    /**
     * Initialises the connection to the MongoDB server
     */
    connect(): Promise<void>;
    /**
     * Disconnects from the MongoDB server
     */
    disconnect(): Promise<void>;
    private _getCollectionName;
    private _convertOrderBy;
    private _getNextAutoNumberValue;
    /**
     * @private
     */
    create<T extends IModel>(manager: ModelManager, model: T, params: ICreateParams, result: ModelOperationResult<T, ICreateMeta>): Promise<ModelOperationResult<T, ICreateMeta>>;
    /**
     * @private
     */
    update<T extends IModel>(manager: ModelManager, model: T, params: IUpdateParams, result: ModelOperationResult<T, IUpdateMeta>): Promise<ModelOperationResult<T, IUpdateMeta>>;
    /**
     * @private
     */
    remove<T extends IModel>(manager: ModelManager, model: T, params: IRemoveParams, result: ModelOperationResult<T, IRemoveMeta>): Promise<ModelOperationResult<T, IRemoveMeta>>;
    /**
     * @private
     */
    read<T extends IModel>(manager: ModelManager, model: new () => T, params: IReadParams, result: ModelOperationResult<T, IReadMeta>): Promise<ModelOperationResult<T, IReadMeta>>;
    /**
     * @private
     */
    exec<R>(manager: ModelManager, model: IModel, params: IExecParams, result: ModelOperationResult<R, IExecMeta>): Promise<ModelOperationResult<R, IExecMeta>>;
}
