"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var rev_models_1 = require("rev-models");
var mongodb_1 = require("mongodb");
var utils_1 = require("rev-models/lib/backends/utils");
var query_1 = require("./query");
/**
 * The MongoDB Collection used to store the next @AutoNumberField values
 */
exports.AUTONUMBER_COLLECTION_NAME = '__revjs_autonumber';
/**
 * A RevJS Backend class that stores model data in MongoDB
 *
 * Usage example:
 *
 * ```ts
 * [[include:examples/src/using_backends/using_a_mongodb_backend.ts]]
 * ```
 */
var MongoDBBackend = /** @class */ (function () {
    /**
     * Backend constructor.
     *
     * **NOTE:** You must call the **[[MongoDBBackend.connect]]** method after constructing
     * the backend before attempting to store or retrieve any data.
     * @param config MongoDB Connection Information
     */
    function MongoDBBackend(config) {
        this.config = config;
    }
    /**
     * Initialises the connection to the MongoDB server
     */
    MongoDBBackend.prototype.connect = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        _a = this;
                        return [4 /*yield*/, mongodb_1.MongoClient.connect(this.config.url, this.config.options)];
                    case 1:
                        _a.client = _b.sent();
                        this.db = this.client.db(this.config.dbName);
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Disconnects from the MongoDB server
     */
    MongoDBBackend.prototype.disconnect = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                if (this.client) {
                    this.client.close();
                    this.client = null;
                }
                return [2 /*return*/];
            });
        });
    };
    MongoDBBackend.prototype._getCollectionName = function (meta) {
        return meta.name;
    };
    MongoDBBackend.prototype._convertOrderBy = function (orderBy) {
        var sortDocument = {};
        if (orderBy instanceof Array) {
            orderBy.forEach(function (field) {
                var tokens = field.split(' ');
                var order = tokens[1] == 'desc' ? -1 : 1;
                sortDocument[tokens[0]] = order;
            });
        }
        return sortDocument;
    };
    MongoDBBackend.prototype._getNextAutoNumberValue = function (modelName, fieldName) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var sequenceName, res;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        sequenceName = modelName + "__" + fieldName;
                        return [4 /*yield*/, this.db.collection(exports.AUTONUMBER_COLLECTION_NAME)
                                .findOneAndUpdate({ _id: sequenceName }, { $inc: { nextValue: 1 } }, { upsert: true, returnOriginal: false })];
                    case 1:
                        res = _a.sent();
                        return [2 /*return*/, res.value.nextValue];
                }
            });
        });
    };
    /**
     * @private
     */
    MongoDBBackend.prototype.create = function (manager, model, params, result) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var meta, document, fieldList, _i, fieldList_1, fieldName, field, _a, _b, value, colName, createResult;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        meta = manager.getModelMeta(model);
                        document = {};
                        fieldList = Object.keys(meta.fieldsByName);
                        _i = 0, fieldList_1 = fieldList;
                        _c.label = 1;
                    case 1:
                        if (!(_i < fieldList_1.length)) return [3 /*break*/, 5];
                        fieldName = fieldList_1[_i];
                        field = meta.fieldsByName[fieldName];
                        if (!field.options.stored) return [3 /*break*/, 4];
                        if (!(field instanceof rev_models_1.fields.AutoNumberField
                            && typeof model[fieldName] == 'undefined')) return [3 /*break*/, 3];
                        _a = document;
                        _b = fieldName;
                        return [4 /*yield*/, this._getNextAutoNumberValue(meta.name, fieldName)];
                    case 2:
                        _a[_b] = _c.sent();
                        return [3 /*break*/, 4];
                    case 3:
                        if (typeof model[fieldName] != 'undefined') {
                            value = field.toBackendValue(manager, model[fieldName]);
                            if (typeof value != 'undefined') {
                                document[fieldName] = value;
                            }
                        }
                        _c.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 1];
                    case 5:
                        colName = this._getCollectionName(meta);
                        return [4 /*yield*/, this.db.collection(colName).insertOne(document)];
                    case 6:
                        createResult = _c.sent();
                        if (createResult.insertedCount != 1) {
                            throw new Error('mongodb insert failed'); // TODO: Something nicer
                        }
                        result.result = manager.hydrate(meta.ctor, document);
                        return [2 /*return*/, result];
                }
            });
        });
    };
    /**
     * @private
     */
    MongoDBBackend.prototype.update = function (manager, model, params, result) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var meta, mongoQuery, fieldUpdates, colName, updateResult;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!params.where) {
                            throw new Error("update() requires the 'where' option to be set.");
                        }
                        meta = manager.getModelMeta(model);
                        mongoQuery = query_1.convertQuery(manager, meta.ctor, params.where);
                        fieldUpdates = {};
                        params.fields.forEach(function (fieldName) {
                            var field = meta.fieldsByName[fieldName];
                            if (field.options.stored && typeof model[fieldName] != 'undefined') {
                                var value = field.toBackendValue(manager, model[fieldName]);
                                if (typeof value != 'undefined') {
                                    fieldUpdates[fieldName] = value;
                                }
                            }
                        });
                        colName = this._getCollectionName(meta);
                        return [4 /*yield*/, this.db.collection(colName).updateMany(mongoQuery, { $set: fieldUpdates })];
                    case 1:
                        updateResult = _a.sent();
                        result.setMeta({ totalCount: updateResult.matchedCount });
                        return [2 /*return*/, result];
                }
            });
        });
    };
    /**
     * @private
     */
    MongoDBBackend.prototype.remove = function (manager, model, params, result) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var meta, mongoQuery, colName, removeResult;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!params.where) {
                            throw new Error('remove() requires the \'where\' option to be set');
                        }
                        meta = manager.getModelMeta(model);
                        mongoQuery = query_1.convertQuery(manager, meta.ctor, params.where);
                        colName = this._getCollectionName(meta);
                        return [4 /*yield*/, this.db.collection(colName).deleteMany(mongoQuery)];
                    case 1:
                        removeResult = _a.sent();
                        result.setMeta({ totalCount: removeResult.deletedCount });
                        return [2 /*return*/, result];
                }
            });
        });
    };
    /**
     * @private
     */
    MongoDBBackend.prototype.read = function (manager, model, params, result) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var meta, colName, mongoQuery, records, primaryKeyValues, foreignKeyValues, rawValues, relatedFieldNames, related, relatedModelInstances, relatedModelListInstances, _i, _a, instance, _b, relatedFieldNames_1, fieldName, field;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        meta = manager.getModelMeta(model);
                        colName = this._getCollectionName(meta);
                        mongoQuery = query_1.convertQuery(manager, model, params.where);
                        return [4 /*yield*/, this.db.collection(colName)
                                .find(mongoQuery)
                                .sort(this._convertOrderBy(params.orderBy))
                                .skip(params.offset)
                                .limit(params.limit)
                                .toArray()];
                    case 1:
                        records = _c.sent();
                        primaryKeyValues = [];
                        foreignKeyValues = {};
                        rawValues = [];
                        relatedFieldNames = utils_1.getOwnRelatedFieldNames(params.related);
                        result.results = [];
                        records.forEach(function (record) {
                            var modelInstance = manager.hydrate(meta.ctor, record);
                            result.results.push(modelInstance);
                            if (meta.primaryKey) {
                                primaryKeyValues.push(modelInstance[meta.primaryKey]);
                            }
                            // TODO: Refactor this into a method shared with InMemoryBackend (and future backends)
                            if (relatedFieldNames) {
                                for (var _i = 0, relatedFieldNames_2 = relatedFieldNames; _i < relatedFieldNames_2.length; _i++) {
                                    var fieldName = relatedFieldNames_2[_i];
                                    var field = meta.fieldsByName[fieldName];
                                    var keyValue = record[fieldName];
                                    if (field instanceof rev_models_1.fields.RelatedModelField) {
                                        if (!(fieldName in foreignKeyValues)) {
                                            foreignKeyValues[fieldName] = [];
                                        }
                                        if (typeof keyValue != 'undefined' && keyValue !== null) {
                                            modelInstance[fieldName] = keyValue;
                                            if (foreignKeyValues[fieldName].indexOf(keyValue) == -1) {
                                                foreignKeyValues[fieldName].push(keyValue);
                                            }
                                        }
                                        else {
                                            modelInstance[fieldName] = null;
                                        }
                                    }
                                }
                            }
                            if (params.rawValues) {
                                var rawValueObj = {};
                                for (var _a = 0, _b = params.rawValues; _a < _b.length; _a++) {
                                    var fieldName = _b[_a];
                                    rawValueObj[fieldName] = record[fieldName];
                                }
                                rawValues.push(rawValueObj);
                            }
                        });
                        if (!relatedFieldNames) return [3 /*break*/, 3];
                        return [4 /*yield*/, Promise.all([
                                utils_1.getRelatedModelInstances(manager, meta, foreignKeyValues, params),
                                utils_1.getRelatedModelListInstances(manager, meta, primaryKeyValues, params)
                            ])];
                    case 2:
                        related = _c.sent();
                        relatedModelInstances = related[0], relatedModelListInstances = related[1];
                        for (_i = 0, _a = result.results; _i < _a.length; _i++) {
                            instance = _a[_i];
                            for (_b = 0, relatedFieldNames_1 = relatedFieldNames; _b < relatedFieldNames_1.length; _b++) {
                                fieldName = relatedFieldNames_1[_b];
                                field = meta.fieldsByName[fieldName];
                                if (field instanceof rev_models_1.fields.RelatedModelField) {
                                    if (instance[fieldName] !== null
                                        && relatedModelInstances[fieldName]
                                        && relatedModelInstances[fieldName][instance[fieldName]]) {
                                        instance[fieldName] = relatedModelInstances[fieldName][instance[fieldName]];
                                    }
                                    else {
                                        instance[fieldName] = null;
                                    }
                                }
                                else if (field instanceof rev_models_1.fields.RelatedModelListField) {
                                    if (relatedModelListInstances[fieldName]
                                        && relatedModelListInstances[fieldName][instance[meta.primaryKey]]) {
                                        instance[fieldName] = relatedModelListInstances[fieldName][instance[meta.primaryKey]];
                                    }
                                    else {
                                        instance[fieldName] = [];
                                    }
                                }
                            }
                        }
                        _c.label = 3;
                    case 3:
                        if (params.rawValues) {
                            result.setMeta({ rawValues: rawValues });
                        }
                        result.setMeta({
                            offset: params.offset,
                            limit: params.limit,
                            totalCount: result.results.length
                        });
                        return [2 /*return*/, result];
                }
            });
        });
    };
    /**
     * @private
     */
    MongoDBBackend.prototype.exec = function (manager, model, params, result) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, Promise.reject(new Error('Not yet implemented'))];
            });
        });
    };
    return MongoDBBackend;
}());
exports.MongoDBBackend = MongoDBBackend;
//# sourceMappingURL=backend.js.map